--[[
    Coconuts by LordFarlander
    Version 2.0
    $Revision: 77774 $

    Used to be called Mount.
]]--

--[[
Copyright (c) 2008, LordFarlander
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
]]--

local L = LibStub( "AceLocale-3.0" ):GetLocale( "Coconuts" );

_G["BINDING_NAME_CLICK " .. L["Coconuts"] .. "_Button:LeftButton"] = L["Summon Mount"];
BINDING_HEADER_COCONUTSHEADER = L["Coconuts"];

if( not LibStub( "LibSpecialEvents-Mount-4.0", true ) ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibSpecialEvents-Mount-4.0" ) );
    return;
end--if

if( not LibStub( "LibSpecialEvents-LearnSpell-4.0", true ) ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibSpecialEvents-Mount-4.0" ) );
    return;
end--if

if( not LibStub( "LibSpecialEvents-Skill-4.0", true ) ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibSpecialEvents-Skill-4.0" ) );
    return;
end--if

local LibLordFarlander = LibStub( "LibLordFarlander-2.0", true );
if( not LibLordFarlander ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-2.0" ) );
    return;
end--if

local ItemList = LibStub( "LibLordFarlander-ItemList-2.0", true );
if( not ItemList ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-ItemList-2.0" ) );
    return;
end--if

if( not LibStub( "LibLordFarlander-UI-2.0", true ) ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-UI-2.0" ) );
    return;
end--if

local PetAndMountDatabase = LibStub( "LibPetAndMountDatabase-1.1", true );
if( not PetAndMountDatabase ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibPetAndMountDatabase-1.1" ) );
    return;
end--if

Coconuts = LibStub( "AceAddon-3.0" ):NewAddon( L["Coconuts"], "AceEvent-3.0", "AceTimer-3.0", "AceConsole-3.0" );
Coconuts.version = tonumber( string.match( "$Rev: 77774 $", "(%d+)" ) );

Coconuts.DEBUG  = false;            --true = Debugging Mode
Coconuts.dprint = LibLordFarlander.dprint;
Coconuts.button = nil;
Coconuts.ui = { currentZone = "", currentFavorite = 0, };

--DO NOT CHANGE THESE VALUES
Coconuts.Ground = { Best = "", NextBest = "", Passenger = {}, Sets = {}, };
Coconuts.Flying = { Best = "", NextBest = "", Passenger = {}, Sets = {}, };

Coconuts.CurrentGround     = nil; --Ground mount on the Button
Coconuts.CurrentFlying     = nil; --Flying mount on the Button
Coconuts.CurrentMount      = nil; --Current mount
Coconuts.CurrentMountItem  = nil; --Current mount modifier item
Coconuts.NextCurrentGround = nil; --Ground mount on the Button
Coconuts.NextCurrentFlying = nil; --Flying mount on the Button
Coconuts.TempleOfAhnQiraj  = nil; --Coconuts allowed in Temple Of Ahn'Qiraj
Coconuts.ReadyToInitialize = false;
Coconuts.Mounted           = false;
Coconuts.Dirty             = false;
Coconuts.Mode              = "";
Coconuts.RidingSkill       = 0;
Coconuts.Loaded            = false;

function Coconuts:OnInitialize()
    local AceConfigReg = LibStub( "AceConfigRegistry-3.0" );
    local AceConfigDialog = LibStub( "AceConfigDialog-3.0" );
    local defaults = {
            profile = {
                FavorFlying = true,
                FavorPassenger = true,
                ShowModifierItems = true,
                SuggestOnlySoulbound = true,
                FlyingModifier = "Alt",
                NextBestModifier = "Shift",
                PassengerModifier = "Control",
                MountItemModifier = "Shift",
                -- List of favorite mounts, these will be picked more often
                -- stored as [itemID]=true
                Favorites = {},
                -- List of mounts per zone, if the mounts are in the player's bags or spells they will be picked
                -- when in the indicated zone
                -- stored as ["Localized zone name"]=itemID
                ZoneFavorites = {},
                Button = LibStub( "LibLordFarlander-UI-2.0" ).GetButtonSettingsDefaults(),
            },
        };
    local buttonOptions = {
        type = "group",
        handler = self,
        name = L["Button"],
        desc = L["Options relating to the button."],
        args = LibStub( "LibLordFarlander-UI-2.0" ).GetButtonOptionArgs( 1, {
            toggleButton = L["toggleButton"],
            ["Show button"] = L["Show button"],
            ["Show/hide the button."] = L["Show/hide the button."],
            autoHideButton = L["autoHideButton"],
            ["Automatically hide button"] = L["Automatically hide button"],
            ["Hide the button automatically?"] = L["Hide the button when you can't mount?"],
            lockButton = L["lockButton"],
            ["Lock button position"] = L["Lock button position"],
            ["Lock/unlock the position of the button."] = L["Lock/unlock the position of the button."],
            ["Alpha of button"] = L["Alpha of button"],
            ["Sets the alpha value that the button has normally."] = L["Sets the alpha value that the button has normally."],
            fadeOut = L["fadeOut"],
            ["Fade out"] = L["Fade out"],
            ["Should the button fade out?"] = L["Fade the button out when the cursor is not over it?"],
            fadeOutInsteadHide = L["fadeOutInsteadHide"],
            ["Fade out instead of hiding"] = L["Fade out instead of hiding"],
            ["Should the button fade out instead of hiding?"] = L["Fade the button out instead of hiding it?"],
            fadeOutCancelInCombat = L["fadeOutCancelInCombat"],
            ["Cancel fade out if in combat"] = L["Cancel fade out if in combat"],
            ["Should the button not fade out if you're in combat?"] = L["Should the button not fade out if you're in combat?"],
            ["Cancel fade out if holding Shift"] = L["Cancel fade out if holding Shift"],
            ["Should the button not fade out if you're holding Shift?"] = L["Should the button not fade out if you're holding Shift?"],
            ["Cancel fade out if holding Ctrl"] = L["Cancel fade out if holding Ctrl"],
            ["Should the button not fade out if you\re holding Ctrl?"] = L["Should the button not fade out if you're holding Ctrl?"],
            ["Cancel fade out if holding Alt"] = L["Cancel fade out if holding Alt"],
            ["Should the button not fade out if you're holding Alt?"] = L["Should the button not fade out if you're holding Alt?"],
            ["Fade out time"] = L["Fade out time"],
            ["Sets how long the button takes to fade out."] = L["Sets how long the button takes to fade out."],
            ["Fade out delay"] = L["Fade out delay"],
            ["Sets the delay before the button starts to fade."] = L["Sets the delay before the button starts to fade."],
            ["Alpha after fadeout"] = L["Alpha after fadeout"],
            ["Sets the alpha value that the button has after faded out."] = L["Sets the alpha value that the button has after faded out."],
            } ),
    };
    local favoritesOptions = {
        type = "group",
        name = L["Favorites"],
        desc = L["Mounts that should be picked more often if you have them."],
        handler = self,
        args = {
            zones = {
                type = "select",
                order = 1,
                name = L["Favorites"],
                desc = L["Mounts that should be picked more often if you have them."],
                width = "full",
                values = function( info )
                             local Favorites = {};
                             local firstItem = 0;

                             for itemID, _ in pairs( info.handler.db.profile.Favorites ) do
                                 local sName, sItemTexture = "", nil;

                                 if( itemID < 0 ) then
                                     sName, _, sItemTexture, _, _, _, _, _, _ = GetSpellInfo( itemID * -1 );
                                 else
                                     sName, _, _, _, _, _, _, _, _, sItemTexture = GetItemInfo( itemID );
                                 end--if
                                 if( sItemTexture ) then
                                     Favorites[itemID] = ("|T%s:0|t %s"):format( sItemTexture, sName );
                                 else
                                     Favorites[itemID] = sName;
                                 end--if
                                 if( firstItem == 0 ) then
                                     firstItem = itemID;
                                 end--if
                             end--for
                             if( ( not info.handler.db.profile.Favorites[info.handler.ui.currentFavorite] ) or ( info.handler.ui.currentFavorite == 0 ) ) then
                                 info.handler.ui.currentFavorite = firstItem;
                             end--if
                             return Favorites;
                         end,
                get = function( info )
                          return info.handler.ui.currentFavorite;
                      end,
                set = function( info, val )
                          info.handler.ui.currentFavorite = val;
                      end,
            },
            mount = {
                type = "input",
                order = 2,
                name = L["Mount"],
                desc = L["Enter the name or ID (item or spell) of a mount to add to the list then press Enter or click Okay."],
                width = "full",
                get = function( info )
                          return info.handler.db.profile.Favorites[info.handler.ui.currentFavorite] or "";
                      end,
                set = function( info, val )
                          local itemLink = LibLordFarlander.GetItemLink( val );
                          local spellLink = GetSpellLink( val, "" );
                          
                          if( itemLink and PetAndMountDatabase:ItemInSet( itemLink, "Mounts" ) ) then
                              local itemID = LibLordFarlander.GetItemIDFromLink( itemLink );

                              info.handler.db.profile.Favorites[itemID] = true;
                              info.handler.ui.currentFavorite = itemID;
                          elseif( spellLink ) then
                              local itemID = LibLordFarlander.GetItemIDFromLink( spellLink ) * -1;

                              info.handler.db.profile.Favorites[itemID] = true;
                              info.handler.ui.currentFavorite = itemID;
                          end--if
                      end,
                validate = function( info, val )
                               local itemLink = LibLordFarlander.GetItemLink( val );
                               local spellLink = GetSpellLink( val, "" );

                               if( itemLink and PetAndMountDatabase:ItemInSet( itemLink, "Mounts" ) ) then
                                   return true;
                               elseif( spellLink and PetAndMountDatabase:SpellInSet( LibLordFarlander.GetItemIDFromLink( spellLink ), "Mounts" ) ) then
                                   return true;
                               end--if
                               if( type( val ) == "number" ) then
                                   return L["Cannot find a mount with itemID or spellID %i!"]:format( val );
                               else
                                   return L["Cannot find a mount named %q!"]:format( val );
                               end--if
                               return false;
                           end,
            },
            delete = {
                type = "execute",
                order = 3,
                name = L["Delete"],
                desc = L["Delete the currently selected mount from the list."],
                func = function( info )
                           info.handler.db.profile.Favorites[info.handler.ui.currentFavorite] = nil;
                           info.handler.ui.currentFavorite = 0;
                       end,
            },
        },
    };
    local zoneOptions =  {
        type = "group",
        name = L["Zone Favorites"],
        desc = L["Set a mount to use when in a certain zone."],
        handler = self,
        args = {
            zones = {
                type = "select",
                order = 1,
                name = L["Zones"],
                desc = L["Set a mount to use when in a certain zone."],
                width = "full",
                values = function( info )
                             local zones = {};
                             local firstItem = 0;
                             local continents = { GetMapContinents() };

                             for continent, _ in pairs( continents ) do
                                 local continentZones = { GetMapZones( continent ) };

                                 for _, zone in pairs( continentZones ) do
                                     local sName, sItemTexture = L["none"], nil;
                                     local itemID = info.handler.db.profile.ZoneFavorites[zone];

                                     if( itemID ) then
                                         if( itemID < 0 ) then
                                             sName, _, sItemTexture, _, _, _, _, _, _ = GetSpellInfo( itemID * -1 );
                                         else
                                             sName, _, _, _, _, _, _, _, _, sItemTexture = GetItemInfo( itemID );
                                         end--if
                                     end--if
                                     if( not sName ) then
                                         sName = L["none"];
                                         info.handler.db.profile.ZoneFavorites[zone] = nil;
                                     end--if
                                     if( sItemTexture ) then
                                         zones[zone] = ("%s [|T%s:0|t %s]"):format( zone, sItemTexture, sName );
                                     else
                                         zones[zone] = ("%s [%s]"):format( zone, sName );
                                     end--if
                                 end--for
                             end--for
                             if( #zones ) then
                                 if( info.handler.ui.currentZone == "" ) then
                                     info.handler.ui.currentZone = GetRealZoneText();
                                 end--if
                             else
                                 info.handler.ui.currentZone = "";
                             end--if                                     
                             return zones;
                         end,
                get = function( info )
                          return info.handler.ui.currentZone;
                      end,
                set = function( info, val )
                          info.handler.ui.currentZone = val;
                      end,
            },
            mount = {
                type = "input",
                order = 2,
                name = L["Mount"],
                desc = L["Enter the name or ID (item or spell) of a mount to add to the list then press Enter or click Okay."],
                width = "full",
                get = function( info )
                          local itemID = info.handler.db.profile.ZoneFavorites[zone];

                          if( itemID ) then
                              if( itemID < 0 ) then
                                  return GetSpellInfo( itemID * -1 ) or "";
                              else
                                  return GetItemInfo( itemID ) or "";
                              end--if
                          end--if
                          return "";
                      end,
                set = function( info, val )
                          local itemLink = LibLordFarlander.GetItemLink( val );
                          local spellLink = GetSpellLink( val, "" );
                          
                          if( itemLink and PetAndMountDatabase:ItemInSet( itemLink, "Mounts" ) ) then
                              info.handler.db.profile.ZoneFavorites[info.handler.ui.currentZone] = LibLordFarlander.GetItemIDFromLink( itemLink );
                          elseif( ( val ~= "" ) and spellLink ) then
                              info.handler.db.profile.ZoneFavorites[info.handler.ui.currentZone] = LibLordFarlander.GetItemIDFromLink( spellLink ) * -1;
                          else
                              info.handler.db.profile.ZoneFavorites[info.handler.ui.currentZone] = nil;
                          end--if
                          info.handler:SetButton( info.handler:SetButtonBasedOnModifiers() );
                      end,
                validate = function( info, val )
                               local itemLink = LibLordFarlander.GetItemLink( val );
                               local spellLink = GetSpellLink( val, "" );

                               if( ( val == "" ) or ( itemLink and PetAndMountDatabase:ItemInSet( itemLink, "Mounts" ) ) ) then
                                   return true;
                               elseif( spellLink and PetAndMountDatabase:SpellInSet( LibLordFarlander.GetItemIDFromLink( spellLink ), "Mounts" ) ) then
                                   return true;
                               end--if
                               if( type( val ) == "number" ) then
                                   return L["Cannot find a mount with itemID or spellID %i!"]:format( val );
                               else
                                   return L["Cannot find a mount named %q!"]:format( val );
                               end--if
                               return false;
                           end,
            },
        },
    };
    local modifierKeys = { ["Alt"] = L["Alt"], ["Control"] = L["Control"], ["Shift"] = L["Shift"], };
    local mainOptions = {
        type = "group",
        handler = self,
        args = {
            debug = {
                type = "toggle",
                hidden = true,
                name = "Debug",
                desc = "Debug mode toggle",
                get = function( info )
                          return info.handler.DEBUG;
                      end,
                set = function( info, val )
                          info.handler.DEBUG = val;
                      end,
            },
            [L["favorFlying"]] = {
                type = "toggle",
                order = 1,
                name = L["Favor flying mounts"],
                desc = L["Toggles if flying mounts should be favored if in an area that allows them."],
                get = function( info )
                          return info.handler.db.profile.FavorFlying;
                      end,
                set = function( info, val )
                          info.handler.db.profile.FavorFlying = val;
                          info.handler:SetButton( info.handler:SetButtonBasedOnModifiers() );
                      end,
            },
            [L["favorPassenger"]] = {
                type = "toggle",
                order = 2,
                name = L["Favor passenger mounts"],
                desc = L["Toggles if mounts that can take a passenger should be favored."],
                get = function( info )
                          return info.handler.db.profile.FavorPassenger;
                      end,
                set = function( info, val )
                          info.handler.db.profile.FavorPassenger = val;
                          info.handler:SetButton( info.handler:SetButtonBasedOnModifiers() );
                      end,
            },
            [L["showModifierItems"]] = {
                type = "toggle",
                order = 3,
                name = L["Show modifier items"],
                desc = L["Toggles if items that modify mounts (such as Fresh Holly) should be selected when holding Ctrl while mounted."],
                get = function( info )
                          return info.handler.db.profile.ShowModifierItems;
                      end,
                set = function( info, val )
                          info.handler.db.profile.ShowModifierItems = val;
                          info.handler:SetButton( info.handler:SetButtonBasedOnModifiers() );
                      end,
            },
            [L["suggestOnlySoulbound"]] = {
                type = "toggle",
                order = 4,
                name = L["Suggest only soulbound mounts"],
                desc = L["Toggles if only mounts that are soulbound to you are to be suggested for use."],
                get = function( info )
                          return Coconuts.db.profile.SuggestOnlySoulbound;
                      end,
                set = function( info, val )
                          info.handler.db.profile.SuggestOnlySoulbound = val;
                          info.handler:SetButton( info.handler:SetButtonBasedOnModifiers() );
                      end,
            },
            flyingModifier = {
                type = "select",
                order = 5,
                name = L["Modifier key for flying mounts"],
                values = modifierKeys,
                cmdHidden = true,
                desc = L["Changes the key that needs to be held down to pick a non flying mount in an area where flying is allowed."],
                width = "full",
                get = function( info )
                          return info.handler.db.profile.FlyingModifier;
                      end,
                set = function( info, val )
                          info.handler.db.profile.FlyingModifier = val;
                      end,
            },
            passengerModifier = {
                type = "select",
                order = 6,
                name = L["Modifier key for passenger mounts"],
                values = modifierKeys,
                cmdHidden = true,
                desc = L["Changes the key that needs to be held down to pick a non passenger mount."],
                width = "full",
                get = function( info )
                          return info.handler.db.profile.PassengerModifier;
                      end,
                set = function( info, val )
                          info.handler.db.profile.PassengerModifier = val;
                      end,
            },
            mountItemModifier = {
                type = "select",
                order = 7,
                name = L["Modifier key for modifier items"],
                values = modifierKeys,
                cmdHidden = true,
                desc = L["Changes the key that needs to be held down to pick a modifer item while mounted."],
                width = "full",
                get = function( info )
                          return info.handler.db.profile.MountItemModifier;
                      end,
                set = function( info, val )
                          info.handler.db.profile.MountItemModifier = val;
                      end,
            },
            nextBestModifier = {
                type = "select",
                order = 8,
                name = L["Modifier key for next slowest mount"],
                values = modifierKeys,
                cmdHidden = true,
                desc = L["Changes the key that needs to be held down to pick your next slowest mount."],
                width = "full",
                get = function( info )
                          return info.handler.db.profile.NextBestModifier;
                      end,
                set = function( info, val )
                          info.handler.db.profile.NextBestModifier = val;
                      end,
            },
            [L["list"]] = {
                type = "execute",
                order = 9,
                name = L["List all mounts"],
                desc = L["Show a list with all your mounts."],
                func = "ListMounts"
            },
            [L["refresh"]] = {
                type = "execute",
                order = 10,
                name = L["Refresh mount list"],
                desc = L["Force a refresh of the list of all your mounts."],
                func = function( info )
                           info.handler:GetPlayersMounts();
                           info.handler:SetButton( info.handler:SetButtonBasedOnModifiers() );
                       end,
            },
        },
        name = L["Coconuts"],
        desc = L["Smart mount randomization on a button! Empty coconut halves not included."],
        icon = "Interface\\AddOns\\Coconuts\\Img\\Coconuts",
    };

    self.db = LibStub( "AceDB-3.0" ):New( "CoconutsDB", defaults );
    self.db.RegisterCallback( self, "OnProfileChanged" );

    LibStub( "AceConfig-3.0" ):RegisterOptionsTable( L["Coconuts"], mainOptions, { L["coconuts"], L["ccn"] } );
    AceConfigReg:RegisterOptionsTable( L["Coconuts Button"], buttonOptions );
    AceConfigReg:RegisterOptionsTable( L["Coconuts Zone Favorites"], zoneOptions );
    AceConfigReg:RegisterOptionsTable( L["Coconuts Favorites"], favoritesOptions );
    AceConfigReg:RegisterOptionsTable( L["Coconuts Profile"], LibStub( "AceDBOptions-3.0" ):GetOptionsTable( self.db ) );
    AceConfigDialog:AddToBlizOptions( L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( L["Coconuts Button"], L["Button"], L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( L["Coconuts Zone Favorites"], L["Zone Favorites"], L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( L["Coconuts Favorites"], L["Favorites"], L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( L["Coconuts Profile"], L["Profile"], L["Coconuts"] );
end--Coconuts:OnInitialize()

function Coconuts:OnProfileReset()
    self:OnProfileChanged();
end--Coconuts:OnProfileReset()

function Coconuts:OnProfileChanged()
    LibStub( "LibLordFarlander-UI-2.0" ).OnProfileChanged( self );
    self:SetButton( self:SetButtonBasedOnModifiers() );
end--Coconuts:OnProfileChanged()

function Coconuts:OnEnable()
    --Register the Events
    local LibSpecialEvents_Mount = LibStub("LibSpecialEvents-Mount-4.0");
    local LibSpecialEvents_Skill = LibStub("LibSpecialEvents-Skill-4.0");
    local LibUI = LibStub( "LibLordFarlander-UI-2.0" );

    self:RegisterEvent( "BAG_UPDATE", "Update" );
    self:RegisterEvent( "ZONE_CHANGED_NEW_AREA",
        function()
            if( Coconuts.DEBUG and LibLordFarlander.IsInFlyableZone() ) then
                Coconuts:dprint( "Coconuts: In flyable zone" );
            end--if
            Coconuts:SetMountsForZone();
            Coconuts:SetButton( Coconuts:SetButtonBasedOnModifiers() );
        end );
    self:RegisterEvent( "MODIFIER_STATE_CHANGED",
        function()
            Coconuts:SetButton( Coconuts:SetButtonBasedOnModifiers() );
        end );
    self:RegisterEvent( "PLAYER_REGEN_DISABLED",
        function()
            if( Coconuts.db.profile.Button.AutoHideButton ) then
                Coconuts.button:HideEx();
            end--if
        end );
    self:RegisterEvent( "PLAYER_REGEN_ENABLED",
        function()
            if( Coconuts.db.profile.Button.AutoHideButton ) then
                Coconuts.button:ShowEx();
            end--if
            Coconuts.button:OutOfCombatReset();
            Coconuts:SetButton( Coconuts:SetButtonBasedOnModifiers() );
        end );
    self:RegisterEvent( "PLAYER_ALIVE",
        function()
            if( not Coconuts.Loaded ) then
                Coconuts.Loaded = true;
                Coconuts.RidingSkill = LibSpecialEvents_Skill:GetSkillLevel( L["Riding"] );
                --Create the Coconuts Inventory
                Coconuts:GetPlayersMounts();
                --Set a Mount
                Coconuts:ChooseRandomMounts();
            end--if
        end );

    LibSpecialEvents_Mount.RegisterCallback( self, "Mounted",
        function()
            if( not Coconuts.Mounted ) then
                Coconuts.Mounted = true;
                Coconuts:ChooseRandomMounts();
            end--if
        end );
    LibSpecialEvents_Mount.RegisterCallback( self, "Dismounted",
        function()
            if( Coconuts.Mounted ) then
                Coconuts.Mounted = false;
                Coconuts:SetButton( Coconuts:SetButtonBasedOnModifiers() );
            end--if
        end );
    LibStub("LibSpecialEvents-LearnSpell-4.0").RegisterCallback( self, "LearnedSpell",
        function( _, spell )
            local spellID = LibLordFarlander.GetItemIDFromLink( GetSpellLink( spell ) );

            if( spellID ~= nil ) then
                if( PetAndMountDatabase:SpellInSet( spellID, "Spells.Mounts" ) ) then
                    Coconuts:Update();
                    return;
                end--if
            end--if
        end );
    LibSpecialEvents_Skill.RegisterCallback( self, "LearnedSkill",
        function( _, skillName )
            if( skillName == L["Riding"] ) then
                Coconuts.RidingSkill = 75;
                Coconuts:SetButton( Coconuts:SetButtonBasedOnModifiers() );
            end--if
        end );
    LibSpecialEvents_Skill.RegisterCallback( self, "SkillLeveled",
        function( _, skillName, level )
            if( skillName == L["Riding"] ) then
                Coconuts.RidingSkill = level;
                Coconuts:SetButton( Coconuts:SetButtonBasedOnModifiers() );
            end--if
        end );

    LibUI.SetupForUI( self );
    self.button = LibUI.CreateButton( self, L["Coconuts"], "Interface\\AddOns\\Coconuts\\Img\\Coconuts", self.db.profile.ButtonGroup, self.SkinChanged, self );
    ItemList.AttachToButton( self.button );
    self.button:SetMovable( not self.db.profile.Button.LockButton );

    --Check if upgrading from Mount
    if( GetBindingKey( "BINDING_NAME_CLICK Mount_Button:LeftButton" ) ) then
        SetBindingClick( GetBindingKey( "BINDING_NAME_CLICK Mount_Button:LeftButton" ), self.button:GetName(), "LeftButton" );
        SetBindingClick( nil, "Mount_Button", "LeftButton" );
    end--if

    --Upgrade check
    if( self.db.profile.LastVersion ~= self.version ) then
        self.db.profile.LastVersion = self.version;
    end--if

    self.Dirty = true
end--Coconuts:OnEnable();

function Coconuts:AutoHideChange()
    self:SetButton( self:SetButtonBasedOnModifiers() );
end--Coconuts:AutoHideChange()

function Coconuts:OnUpdate( elapsed )
    if( self.Dirty ) then
        --Update the list of mount items if needed
        self.Dirty = false;        
        if( self:HasMountListChanged() ) then
            self:ChooseRandomMounts();
        end--if
        self:SetButton( self:SetButtonBasedOnModifiers() );
    end--if
end--Coconuts:OnUpdate()

function Coconuts:Update()
    self.Dirty = true;
    if( not self.button:IsVisible() ) then
        self:ScheduleTimer( "OnUpdate", 0 );
    end--if
end--Coconuts:Update()

function Coconuts.IsModifierDown( modname )
  return _G[("Is%sKeyDown"):format( modname )]();
end--Coconuts:IsModifierDown( modname )

function Coconuts:SetButtonBasedOnModifiers()
    local profile = self.db.profile;
    local IsFlyingModifierDown = self.IsModifierDown( profile.FlyingModifier );
    local IsPassengerModifierDown = self.IsModifierDown( profile.PassengerModifier );
    local IsMountItemModifierDown = self.IsModifierDown( profile.MountItemModifier );
    local IsNextBestModifierDown = self.IsModifierDown( profile.NextBestModifier );

    if( self.Mounted and self.db.profile.ShowModifierItems and IsMountItemModifierDown and self.CurrentMountItem ) then
        self.Mode = "Enchant";
        return self.CurrentMountItem;
    elseif( self.Mounted ) then
        self.Mode = "Dismount";
        return nil;
    elseif( ( GetZoneText() == L["Temple Of Ahn'Qiraj"] ) and self.TempleOfAhnQiraj ) then
        self.Mode = "Mount";
        return self.TempleOfAhnQiraj;
    else
        --Pick a mount
        if( self.db.profile.FavorFlying and LibLordFarlander.IsInFlyableZone() and ( self.NextCurrentFlying or self.CurrentFlying ) ) then
            if( IsFlyingModifierDown and ( self.NextCurrentGround or self.CurrentGround ) ) then
                if( IsNextBestModifierDown and self.NextCurrentGround ) then
                    self.Mode = "Mount";
                    return self.NextCurrentGround;
                else
                    self.Mode = "Mount";
                    return self.CurrentGround;
                end--if
            else
                if( IsNextBestModifierDown and self.NextCurrentFlying ) then
                    self.Mode = "Mount";
                    return self.NextCurrentFlying;
                else
                    self.Mode = self.CurrentFlying and "Mount" or "";
                    return self.CurrentFlying;
                end--if
            end--if
        elseif( LibLordFarlander.IsInFlyableZone() and ( self.NextCurrentFlying or self.CurrentFlying ) ) then
            if( IsFlyingModifierDown or not ( self.NextCurrentGround or self.CurrentGround ) ) then
                if( IsNextBestModifierDown and self.NextCurrentFlying ) then
                    self.Mode = "Mount";
                    return self.NextCurrentFlying;
                else
                    self.Mode = "Mount";
                    return self.CurrentFlying;
                end--if
            else
                if( IsNextBestModifierDown and self.NextCurrentGround ) then
                    self.Mode = "Mount";
                    return self.NextCurrentGround;
                else
                    self.Mode = self.CurrentGround and "Mount" or "";
                    return self.CurrentGround;
                end--if
            end--if
        else
            if( IsNextBestModifierDown ) then
                self.Mode = self.NextCurrentGround and "Mount" or "";
                return self.NextCurrentGround;
            else
                self.Mode = self.CurrentGround and "Mount" or "";
                return self.CurrentGround;
            end--if
        end--if
    end--if
    self.Mode = "";
    return nil;
end--Coconuts:SetButtonBasedOnModifiers()

----------------------------
--+Mount/Inventory Functions--
function Coconuts:GetZoneMount( ZoneList, set )
    if( ZoneList and ( set:len() > 0 ) ) then
        if( ( ( ( ZoneList < 0 ) and LibLordFarlander.HasPlayerSpell( ZoneList * -1 ) ) or ( LibLordFarlander.PlayerItemCount( ZoneList ) > 0 ) ) and
            PetAndMountDatabase:ItemInSet( ZoneList, set ) ) then
            local BagItem = ItemList.FindInBags( ZoneList );
            
            if( ( not self.db.profile.SuggestOnlySoulbound ) or LibLordFarlander.ItemIsSoulbound( BagItem.bag, BagItem.slot ) ) then
                return BagItem;
            end--if
        end--if
    end--if
    return nil;
end--Coconuts:GetZoneMount( ZoneList, PT3set )

function Coconuts:GetRandomFromListWithWeighing( masterset, set, current, favorites )
    if( masterset[set] ) then
      local useset;
      local setmain = masterset[set].Items;

      if( ( #favorites > 0 ) and setmain and ( #setmain > 1 ) ) then
          useset = {};
          LibLordFarlander.CopyTable( useset, setmain );
          for itemID, _ in pairs( favorites ) do
              if( ( ( ( itemID < 0 ) and LibLordFarlander.HasPlayerSpell( itemID * -1 ) ) or ( LibLordFarlander.PlayerItemCount( itemID ) > 0 ) ) and
                  PetAndMountDatabase:ItemInSet( itemID, set ) ) then
                  local BagItem = ItemList.FindInBags( itemID );
                  
                  if( ( not self.db.profile.SuggestOnlySoulbound ) or LibLordFarlander.ItemIsSoulbound( BagItem.bag, BagItem.slot ) ) then
                      table.insert( useset, BagItem );
                  end--if
              end--if
          end--for
      else
          useset = setmain;
      end--if
      return ItemList.GetRandomItemFromList( useset, current );
  end--if
  return nil;
end--Coconuts:GetRandomFromListWithWeighing( masterset, set, current, favorites )

function Coconuts:SetMountsForZone()
    local ZoneList = self.db.profile.ZoneFavorites[GetRealZoneText()];

    if( ZoneList ) then
        self.CurrentGround = self:GetZoneMount( ZoneList, self.Ground.Best ) or self.CurrentGround;
        self.NextCurrentGround = self:GetZoneMount( ZoneList, self.Ground.NextBest ) or self.NextCurrentGround;
        self.CurrentFlying = self:GetZoneMount( ZoneList, self.Flying.Best ) or self.CurrentFlying;
        self.NextCurrentFlying = self:GetZoneMount( ZoneList, self.Flying.NextBest ) or self.NextCurrentFlying;
    end--if
end--Coconuts:SetMountsForZone()

function Coconuts:ChooseRandomMounts()
    local Favorites = self.db.profile.Favorites;
    local GroundSets = self.Ground.Sets;
    local FlyingSets = self.Flying.Sets;

    self.CurrentGround = self:GetRandomFromListWithWeighing( GroundSets, self.Ground.Best, self.CurrentGround, Favorites );
    self.NextCurrentGround = self:GetRandomFromListWithWeighing( GroundSets, self.Ground.NextBest, self.NextCurrentGround, Favorites );
    self.CurrentFlying = self:GetRandomFromListWithWeighing( FlyingSets, self.Flying.Best, self.CurrentFlying, Favorites );
    self.NextCurrentFlying = self:GetRandomFromListWithWeighing( FlyingSets, self.Flying.NextBest, self.NextCurrentFlying, Favorites );

    self:SetMountsForZone();
    self:SetButton( self:SetButtonBasedOnModifiers() );
end--Coconuts:ChooseRandomMounts()

function Coconuts.GetBestSets( sets )
    local Best = nil;
    local BestName = "";
    local NextBest = nil;
    local NextBestName = "";

    for name, set in pairs( sets ) do
        if( Best == nil ) then
            Best = set;
            BestName = name;
        else
            if( set.speed > Best.speed ) then
                if( ( NextBest == nil ) or ( Best.speed > NextBest.speed ) ) then
                    NextBest = Best;
                    NextBestName = BestName;
                end--if
                Best = set;
                BestName = name;
            elseif( ( NextBest == nil ) or ( set.speed > NextBest.speed ) ) then
                NextBest = set;
                NextBestName = name;
            end--if
        end--if
    end--for
    if( not NextBest ) then
        NextBestName = BestName;
    end--if
    return BestName, NextBestName;
end--Coconuts:GetBestSets()

function Coconuts.AddIfMount( ptsets, dest, bag, slot, itemID )
    for _, set in pairs( ptsets ) do
        local speed = PetAndMountDatabase:ItemInSet( itemID, set );

        if( speed ) then
            if( dest[set] == nil ) then
                dest[set] = { Items = {}, ["speed"] = PetAndMountDatabase:GetMountAttributes( speed ) };
            end--if
            ItemList.AddBagItemToList( dest[set].Items, bag, slot );
            return true;
        end--if
    end--for
    return false;
end--Coconuts.AddIfMount( ptsets, dest, bag, slot, itemID )

function Coconuts.GetPlayerSpellMounts( spellDatabase, mountTable )
    local iterator, handle = PetAndMountDatabase:IterateSet( spellDatabase );

    if( iterator and handle ) then
        local spellID, speed = iterator( handle );

        while spellID do
            spellID = spellID * -1;
            if( LibLordFarlander.HasPlayerSpell( spellID ) ) then
                if( mountTable.speed == nil ) then
                    mountTable.speed = PetAndMountDatabase:GetMountAttributes( speed );
                end--if
                ItemList.AddSpellToList( mountTable.Items, spellID );
            end--if
            spellID, speed = iterator( handle );
        end--for
    end--if
end--Coconuts.GetPlayerSpellMounts( spellDatabase, mountTable )

function Coconuts.AddSpells( ptsets, dest )
    for _, set in pairs( ptsets ) do
        local SpellSet = ("%s.Spells"):format( set );

        if( PetAndMountDatabase:HasSet( SpellSet ) ) then
            local destSet = dest[set];

            if( destSet == nil ) then
                destSet = { Items = {}, };
            end--if
            Coconuts.GetPlayerSpellMounts( set, destSet );
            if( ( dest[set] == nil ) and ( destSet.speed ~= nil ) ) then
                dest[set] = destSet;
            end--if
        end--if
    end--for
end--Coconuts.AddSpells( ptsets, dest )

--Gets all your pets from your inventory.
function Coconuts:GetPlayersMounts()
    -- First do ground mountItems, best first
    local opSuggestOnlySoulbound = self.db.profile.SuggestOnlySoulbound;

    self.CurrentMountItem = nil;
    self.Ground = { Best = "", NextBest = "", Passenger = {}, Sets = {}, };
    self.Flying = { Best = "", NextBest = "", Passenger = {}, Sets = {}, };

    local GroundMounts = PetAndMountDatabase:GetSubsets( "Mounts.Ground" );
    local FlyingMounts = PetAndMountDatabase:GetSubsets( "Mounts.Flying" );
    local GroundPassengerMounts = PetAndMountDatabase:GetSubsets( "Mounts.Passenger.Ground" );
    local FlyingPassengerMounts = PetAndMountDatabase:GetSubsets( "Mounts.Passenger.Flying" );
    local GroundSets = self.Ground.Sets;
    local FlyingSets = self.Flying.Sets;
    local GroundPassenger = self.Ground.Passenger;
    local FlyingPassenger = self.Flying.Passenger;

    if( not GroundPassengerMounts ) then
        GroundPassengerMounts = {};
    end--if
    if( not FlyingPassengerMounts ) then
        FlyingPassengerMounts = {};
    end--if
    for bag = 0, NUM_BAG_SLOTS do
        for slot = 1, GetContainerNumSlots( bag ) do
            local itemLink = GetContainerItemLink( bag, slot );

            if( itemLink ) then
                local itemString = LibLordFarlander.GetItemStringFromLink( itemLink );
                local itemID = LibLordFarlander.GetItemIDFromString( itemString );

                if( itemID and ( ( not opSuggestOnlySoulbound ) or LibLordFarlander.ItemIsSoulbound( bag, slot ) ) and self:PlayerMeetsMountRequirements( itemID ) ) then
                    self.AddIfMount( GroundMounts, GroundSets, bag, slot, itemID );
                    self.AddIfMount( FlyingMounts, FlyingSets, bag, slot, itemID );
                    self.AddIfMount( GroundPassengerMounts, GroundPassenger, bag, slot, itemID );
                    self.AddIfMount( FlyingPassengerMounts, FlyingPassenger, bag, slot, itemID );
                    if( ( GetZoneText() == L["Temple Of Ahn'Qiraj"] ) and PetAndMountDatabase:ItemInSet( itemID, "TempleOfAhnQiraj.Mounts" ) ) then
                        self.TempleOfAhnQiraj = ItemList.BuildBagItem( bag, slot );
                    end--if
                end--if
            end--if
        end--for
    end--for

    self.AddSpells( GroundMounts, GroundSets );
    self.AddSpells( FlyingMounts, FlyingSets );

    self.Ground.Best, self.Ground.NextBest = self.GetBestSets( GroundSets );
    self.Flying.Best, self.Flying.NextBest = self.GetBestSets( FlyingSets );
    
    local iterator, handle = PetAndMountDatabase:IterateSet( "Items.MountEnchant" );
    local item, _ = iterator( handle );

    while item do
        if( GetItemCount( item ) > 0 ) then
            self.CurrentMountItem = ItemList.BuildItem( item );
            break;
        end--if
        item, _ = iterator( handle );
    end--for
end--Coconuts:GetPlayersMounts()

function Coconuts:PrintMountList( mountTable, string )
    if( #mountTable > 0 ) then
        self:Print( L["%s mounts:"]:format( string ) );
        for _, mount in pairs( mountTable ) do
            self:Print( ("|T%s:0|t %s"):format( mount:GetTexture(), mount:GetLink() ) );
        end--for
    end--if
end--Coconuts:PrintMountList()

function Coconuts:ListMountType( type, lists )
    local sorted = {};

    for name, _ in pairs( lists ) do
        table.insert( sorted, name );
    end--for
    self.CompareList = lists;
    table.sort( sorted, function( a, b ) return Coconuts.CompareList[a].speed < Coconuts.CompareList[b].speed end );
    self.CompareList = nil;
    for _, name in pairs( sorted ) do
        self:PrintMountList( lists[name].Items, L["%1$i%% speed %2$s"]:format( lists[name].speed, type ) );
    end--for
end--Coconuts.ListMountType( type, list )

--Prints out a list with all your mountItems.
function Coconuts:ListMounts()
    self:Print( L["List of your mounts:"] );

    self:ListMountType( L["ground"], self.Ground.Sets );
    self:ListMountType( L["flying"], self.Flying.Sets );
end--Coconuts:ListMounts()
---Coconuts/Inventory Functions--
----------------------------

-------------------
--+Help Functions--
function Coconuts:PlayerMeetsMountRequirements( itemID )
    local speed = PetAndMountDatabase:ItemInSet( itemID, "Mounts" );
    
    if( speed ) then
        local _, _, _, _, itemMinLevel, _, _, _, _, _ = GetItemInfo( itemID );    
        local _, ridingSkill = PetAndMountDatabase:GetMountAttributes( speed );

        return ( ( itemMinLevel <= UnitLevel( "player" ) ) and ( ridingSkill <= self.RidingSkill ) );
    end--if
    return false;
end--if

function Coconuts:HasMountListChanged()
    local selfBestGround = self.Ground.Best and self.Ground.Sets[self.Ground.Best] or {};
    local selfBestFlying = self.Flying.Best and self.Ground.Sets[self.Flying.Best] or {};
    local selfNextBestGround = self.Ground.NextBest and self.Ground.Sets[self.Ground.NextBest] or {};
    local selfNextBestFlying = self.Flying.NextBest and self.Ground.Sets[self.Flying.NextBest] or {};
    local preGroundCount = #selfBestGround;
    local preFlyingCount = #selfBestFlying;
    local preNextGroundCount = #selfNextBestGround;
    local preNextFlyingCount = #selfNextBestFlying;
    local preBestGround =  {};
    local preBestFlying = {};
    local preNextBestGround =  {};
    local preNextBestFlying = {};

    LibLordFarlander.CopyTable( preBestGround, selfBestGround );
    LibLordFarlander.CopyTable( preBestFlying, selfBestFlying );
    LibLordFarlander.CopyTable( preNextBestGround, selfNextBestGround );
    LibLordFarlander.CopyTable( preNextBestFlying, selfNextBestFlying );

    self:GetPlayersMounts();
    
    selfBestGround = self.Ground.Best and self.Ground.Sets[self.Ground.Best] or {};
    selfBestFlying = self.Flying.Best and self.Ground.Sets[self.Flying.Best] or {};
    selfNextBestGround = self.Ground.NextBest and self.Ground.Sets[self.Ground.NextBest] or {};
    selfNextBestFlying = self.Flying.NextBest and self.Ground.Sets[self.Flying.NextBest] or {};

    -- differences in sizes always means the list has changed
    if( ( preGroundCount ~= #selfBestGround ) or ( preFlyingCount ~= #selfBestFlying ) or ( not self.ReadyToInitialize ) or
        ( preNextGroundCount ~= #selfNextBestGround ) or ( preNextFlyingCount ~= #selfNextBestFlying ) or ( not self.ReadyToInitialize ) ) then
        self.ReadyToInitialize = true;
        return true;
    end
    self.ReadyToInitialize = true;
    return not( LibLordFarlander.CompareTables( preBestGround, selfBestGround ) and LibLordFarlander.CompareTables( preBestFlying, selfBestFlying ) and
        LibLordFarlander.CompareTables( preNextBestGround, selfNextBestGround ) and LibLordFarlander.CompareTables( preNextBestFlying, selfNextBestFlying ) );
end--Coconuts:HasMountListChanged()
---Help Functions--
-------------------

-----------------
--+UI Functions--
function Coconuts:SetGameTooltips()
    local mntButton = self.button;

    if( not self.keyBoundMode ) then
        local mountItem = self.CurrentMount;
        local mntTooltip = mntButton.tooltip;
        local chardb = self.db.char;

        mntTooltip:SetOwner( mntButton, "RIGHT" );
        if( self.Mode == "Dismount" ) then
            if( chardb.ShowModifierItems ) then
                local notUsed = {};
                local used = self.db.profile.MountItemModifier;
                
                if( used ~= "Alt" ) then
                    table.insert( notUsed, L["Alt"] );
                end--if
                if( used ~= "Control" ) then
                    table.insert( notUsed, L["Control"] );
                end--if
                if( used ~= "Shift" ) then
                    table.insert( notUsed, L["Shift"] );
                end--if
                mntTooltip:SetText( L["Dismount (if flying hold down %s or %s and click to dismount)"]:format( notUsed[1], notUsed[2] ) );
            else
                mntTooltip:SetText( L["Dismount (if flying hold down Shift, Control, or Alt and click to dismount)"] );
            end--if
        elseif( ( self.Mode == "Mount" ) or ( self.Mode == "Enchant" ) ) then
            mntTooltip:SetFromListEntry( mountItem );
        else
            mntTooltip:SetText( L["You have no mounts."] );
        end--if
    end--if
    LibStub( "LibLordFarlander-UI-2.0" ).AfterSetTooltip( mntButton );
end--Coconuts:SetGameTooltips()

function Coconuts:SetButton( mountItem )
    local mntButton = self.button;
    local chardb = self.db.profile;

    if( not InCombatLockdown() ) then
        if( chardb.Button.AutoHideButton and mountItem ) then
            mntButton:ShowEx();
        end--if
        if( self.Mode == "Dismount" ) then
            mntButton.cooldown:Hide();
            mntButton:ShowEx();
            mntButton:Enable();
            mntButton:SetAttribute( "*type1", "macro" );
            mntButton:SetAttribute( "macrotext", "/dismount [noflying][mod]" );
            mntButton:SetImage( "Interface\\AddOns\\Coconuts\\Img\\Dismount" );
        elseif( ( self.Mode == "Mount" ) or ( self.Mode == "Enchant" ) ) then
            mntButton:Enable();
            mntButton:SetFromListEntry( mountItem, not LibLordFarlander.CompareTables( self.CurrentMount, mountItem ) );
        else
            if( chardb.Button.AutoHideButton ) then
                mntButton:HideEx();
            end--if
            mntButton.cooldown:Hide();
            mntButton:SetAttribute( "*type1", "item" );
            mntButton:SetAttribute( "*bag1", nil );
            mntButton:SetAttribute( "*slot1", nil );
            mntButton:SetAttribute( "*item1", nil );
            mntButton:SetImage( "Interface\\AddOns\\Coconuts\\Img\\Coconuts" );
        end--if
        self.CurrentMount = mountItem;
    end--if
end--Coconuts:SetButton( mountItem )
---UI Functions--
-----------------

